/*
 * Name:    mpca_epics.c
 *
 * Purpose: This file implements the MpCaPrivate module used to communicate
 *          with external EPICS servers.  Python programs should not invoke
 *          this module directly, but should invoke the MpCa module in
 *          'MpCa.py' instead.
 *
 * Author:  William Lavender
 *
 * Note:    This code requires Python 2.4 or above, due to the use of the
 *          functions PyGILState_Ensure(), PyGILState_Release(), and
 *          PyEval_ThreadsInitialized().
 *
 *---------------------------------------------------------------------------
 *
 * Copyright 2009 Illinois Institute of Technology
 *
 * See the file "LICENSE" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#define MPCA_DEBUG		FALSE

#define MPCA_DEBUG_THREADING	FALSE

#include "Python.h"

#if ( PY_VERSION_HEX < 0x02040000 )
#error MpCA requires Python 2.4 or above.
#endif

#include "mx_util.h"
#include "mx_record.h"
#include "mx_handle.h"
#include "mx_thread.h"
#include "mx_epics.h"

#include "mp.h"
#include "mpca.h"

/* Method table for the MpCaPrivate module. */

static PyMethodDef MpCaMethods[] = {
	{ "init",
			MpCaPrivate_init_cmd,
			METH_VARARGS },

	{ "pend_io",
			MpCaPrivate_pend_io_cmd,
			METH_VARARGS },

	{ "pend_event",
			MpCaPrivate_pend_event_cmd,
			METH_VARARGS },

	{ "flush_io",
			MpCaPrivate_flush_io_cmd,
			METH_VARARGS },

	{ "pv_init",
			MpCaPrivate_pv_init_cmd,
			METH_VARARGS },

	{ "pv_caget",
			MpCaPrivate_pv_caget_cmd,
			METH_VARARGS },

	{ "pv_caput",
			MpCaPrivate_pv_caput_cmd,
			METH_VARARGS },

	{ "pv_add_callback",
			MpCaPrivate_pv_add_callback_cmd,
			METH_VARARGS },

	{ "cb_delete_callback",
			MpCaPrivate_cb_delete_callback_cmd,
			METH_VARARGS },

	{ NULL, NULL }
};

/* The following is the only global variable in MpCaPrivate. */

MX_HANDLE_TABLE *MpCaPrivate_global_pv_handle_table = NULL;

/*---------------------------------------------------------------------------*/

#if MPCA_DEBUG_THREADING
static void
mpca_show_thread_id( const char *label )
{
	char id_buffer[80];

	mx_thread_id_string( id_buffer, sizeof(id_buffer) );

	MX_DEBUG(-2,("%s: %s", label, id_buffer));
}

#  define MPCA_SHOW_THREAD_ID(l)	mpca_show_thread_id(l)
#else
#  define MPCA_SHOW_THREAD_ID(l)
#endif

/*---------------------------------------------------------------------------*/

/* This is MpCaPrivate's module initialization function. */

MX_EXPORT void
initMpCaPrivate( void )
{
	/* Initialize the module. */

	(void) Py_InitModule("MpCaPrivate", MpCaMethods);
}

/*---------------------------------------------------------------------------*/

MX_EXPORT PyObject *
MpCaPrivate_init_cmd( PyObject *self, PyObject *args )
{
#if MPCA_DEBUG_THREADING
	static const char fname[] = "MpCaPrivate_init_cmd()";
#endif

	mx_status_type mx_status;

#if 0
	mx_start_debugger(NULL);
#endif

	MPCA_SHOW_THREAD_ID(fname);

	/* Create a global handle table for the EPICS PVs.  Configure the
	 * table to be incremented in blocks of 100 handles.
	 */

	mx_status = mx_create_handle_table(
					&MpCaPrivate_global_pv_handle_table,
					100, 1 );

	if ( mx_status.code != MXE_SUCCESS )
		return MpError_SetFromMxStatus( NULL, mx_status );

	Py_INCREF(Py_None);

	return Py_None;
}

/*---------------------------------------------------------------------------*/

MX_EXPORT PyObject *
MpCaPrivate_pend_io_cmd( PyObject *self, PyObject *args )
{
#if MPCA_DEBUG_THREADING
	static const char fname[] = "MpCaPrivate_pend_io_cmd()";
#endif

	double timeout;
	int parse_status;
	mx_status_type mx_status;

	MPCA_SHOW_THREAD_ID(fname);

	parse_status = PyArg_ParseTuple( args, "d", &timeout );

	if ( parse_status == 0 )
		return NULL;

	mx_status = mx_epics_pend_io( timeout );

	if ( mx_status.code != MXE_SUCCESS )
		return MpError_SetFromMxStatus( NULL, mx_status );

	Py_INCREF(Py_None);

	return Py_None;
}

/*---------------------------------------------------------------------------*/

MX_EXPORT PyObject *
MpCaPrivate_pend_event_cmd( PyObject *self, PyObject *args )
{
#if MPCA_DEBUG_THREADING
	static const char fname[] = "MpCaPrivate_pend_event_cmd()";
#endif

	double timeout;
	int parse_status;
	mx_status_type mx_status;

	MPCA_SHOW_THREAD_ID(fname);

	parse_status = PyArg_ParseTuple( args, "d", &timeout );

	if ( parse_status == 0 )
		return NULL;

	mx_status = mx_epics_pend_event( timeout );

	if ( mx_status.code != MXE_SUCCESS )
		return MpError_SetFromMxStatus( NULL, mx_status );

	Py_INCREF(Py_None);

	return Py_None;
}

/*---------------------------------------------------------------------------*/

MX_EXPORT PyObject *
MpCaPrivate_flush_io_cmd( PyObject *self, PyObject *args )
{
#if MPCA_DEBUG_THREADING
	static const char fname[] = "MpCaPrivate_flush_io_cmd()";
#endif

	mx_status_type mx_status;

	MPCA_SHOW_THREAD_ID(fname);

	mx_status = mx_epics_flush_io();

	if ( mx_status.code != MXE_SUCCESS )
		return MpError_SetFromMxStatus( NULL, mx_status );

	Py_INCREF(Py_None);

	return Py_None;
}

/*---------------------------------------------------------------------------*/

MX_EXPORT PyObject *
MpCaPrivate_pv_init_cmd( PyObject *self, PyObject *args )
{
	static const char fname[] = "MpCaPrivate_pv_init_cmd()";

	PyObject *result;
	char *pvname;
	MX_EPICS_PV *new_pv;
	long new_pv_handle;
#if 1
	long field_type;
	unsigned long num_elements;
	size_t element_size;
#endif
	int parse_status;
	mx_status_type mx_status;

	MPCA_SHOW_THREAD_ID(fname);

	parse_status = PyArg_ParseTuple( args, "s", &pvname );

	if ( parse_status == 0 )
		return NULL;

#if MPCA_DEBUG
	MX_DEBUG(-2,("%s invoked for PV '%s'.", fname, pvname ));
#endif

	/* Allocate memory for a new MX_EPICS_PV structure. */

	new_pv = malloc( sizeof(MX_EPICS_PV) );

	if ( new_pv == (MX_EPICS_PV *) NULL ) {
		MpError_SetString( PyExc_MemoryError,
	"%s: Unable to allocate memory for an MX_EPICS_PV structure.", fname );

		return NULL;
	}

	/* Initialize the contents of the MX_EPICS_PV structure. */

	mx_epics_pvname_init( new_pv, pvname );

	/* Connect to the PV. */

#if 0
	mx_status = mx_epics_pv_connect( new_pv, FALSE );

	if ( mx_status.code != MXE_SUCCESS ) {
		free( new_pv );
		return MpError_SetFromMxStatus( NULL, mx_status );
	}

	mx_status = mx_epics_pend_io( 1.0 );

	if ( mx_status.code != MXE_SUCCESS ) {
		free( new_pv );
		return MpError_SetFromMxStatus( NULL, mx_status );
	}
#endif

#if 1
	/* Allocate memory for a local Channel Access receive buffer
	 * and store a pointer to it in the PV's application_ptr.
	 */

	field_type = mx_epics_pv_get_field_type( new_pv );

	if ( field_type < 0 ) {
		mx_status = mx_error( MXE_NETWORK_IO_ERROR, fname,
		"Unable to get the field type of EPICS PV '%s'.",
			pvname );

		return MpError_SetFromMxStatus( NULL, mx_status );
	}

	element_size = mx_epics_get_element_size( field_type );

	num_elements = mx_epics_pv_get_element_count( new_pv );

	if ( num_elements == 0 ) {
		mx_status = mx_error( MXE_NETWORK_IO_ERROR, fname,
		"Unable to get the element count of EPICS PV '%s'.",
			pvname );

		return MpError_SetFromMxStatus( NULL, mx_status );
	}

#if MPCA_DEBUG
	MX_DEBUG(-2,("%s: num_elements = %lu, element_size = %lu",
		fname, num_elements, (unsigned long) element_size ));
#endif

	new_pv->application_ptr = malloc( num_elements * element_size );

	if ( new_pv->application_ptr == NULL ) {
		mx_status = mx_error( MXE_OUT_OF_MEMORY, fname,
		"Ran out of memory trying to allocate a receive buffer "
		"with %lu elements of size %lu each for EPICS PV '%s'.",
			num_elements, (unsigned long) element_size,
			pvname );
			
		return MpError_SetFromMxStatus( NULL, mx_status );
	}
#endif

	/* Create a handle for the MX_EPICS_PV pointer. */

	mx_status = mx_create_handle( &new_pv_handle,
					MpCaPrivate_global_pv_handle_table,
					new_pv );

	if ( mx_status.code != MXE_SUCCESS ) {
		free( new_pv );
		return MpError_SetFromMxStatus( NULL, mx_status );
	}

	/* Send the PV handle back to Python. */

	result = Py_BuildValue( "l", new_pv_handle );

#if MPCA_DEBUG
	MX_DEBUG(-2,("%s: new_pv_handle = %ld", fname, new_pv_handle));
#endif

	return result;
}

/*---------------------------------------------------------------------------*/

MX_EXPORT PyObject *
MpCaPrivate_pv_caget_cmd( PyObject *self, PyObject *args )
{
	static const char fname[] = "MpCaPrivate_pv_caget_cmd()";

	PyObject *result = NULL;
	PyObject *item = NULL;
	MX_EPICS_PV *pv;
	void *void_pv, *value_ptr;
	long pv_handle, field_type, remote;
	unsigned long element_count;
	double timeout;
	int i, parse_status, python_status;
	char *char_array = NULL;
	short *short_array = NULL;
	long *long_array = NULL;
	float *float_array = NULL;
	double *double_array = NULL;
	mx_status_type mx_status;

	MPCA_SHOW_THREAD_ID(fname);

#if 0
	mx_breakpoint();
#endif

	parse_status = PyArg_ParseTuple( args, "ldl",
					&pv_handle, &timeout, &remote );

	if ( parse_status == 0 )
		return NULL;

	/* Get the MX_EPICS_PV structure from the handle. */

	mx_status = mx_get_pointer_from_handle( &void_pv,
					MpCaPrivate_global_pv_handle_table,
					pv_handle );

	if ( mx_status.code != MXE_SUCCESS )
		return MpError_SetFromMxStatus( NULL, mx_status );

	pv = void_pv;

	/* Get the PV value from the EPICS server. */

	field_type = mx_epics_pv_get_field_type( pv );

	if ( field_type < 0 ) {
		mx_status = mx_error( MXE_NETWORK_IO_ERROR, fname,
		"Unable to get the native field type for EPICS PV '%s'.",
			pv->pvname );

		return MpError_SetFromMxStatus( NULL, mx_status );
	}
	
	element_count = mx_epics_pv_get_element_count( pv );

	if ( element_count == 0 ) {
		mx_status = mx_error( MXE_NETWORK_IO_ERROR, fname,
		"Unable to get the element count for EPICS PV '%s'.",
			pv->pvname );

		return MpError_SetFromMxStatus( NULL, mx_status );
	}
	
	value_ptr = pv->application_ptr;

#if MPCA_DEBUG
	MX_DEBUG(-2,("%s: remote = %ld", fname, remote));
#endif

	if ( remote ) {
		if ( timeout >= 0.0 ) {
			mx_status = mx_caget_with_timeout( pv,
					field_type, element_count,
					value_ptr, timeout );
		} else {
			mx_status = mx_caget( pv,
					field_type, element_count,
					value_ptr );
		}

		if ( mx_status.code != MXE_SUCCESS )
			return MpError_SetFromMxStatus( NULL, mx_status );
	}

	if ( field_type == MX_CA_STRING ) {
		result = Py_BuildValue( "s", value_ptr );
	} else
	if ( element_count == 1 ) {
		switch( field_type ) {
		case MX_CA_CHAR:
			result = Py_BuildValue( "b", *((char *) value_ptr) );
			break;
		case MX_CA_SHORT:
		case MX_CA_ENUM:
			result = Py_BuildValue( "h", *((short *) value_ptr) );
			break;
		case MX_CA_LONG:
			result = Py_BuildValue( "l", *((long *) value_ptr) );
			break;
		case MX_CA_FLOAT: 
			result = Py_BuildValue( "f", *((float *) value_ptr) );
			break;
		case MX_CA_DOUBLE:
			result = Py_BuildValue( "d", *((double *) value_ptr) );
			break;
		}
	} else {
		/* Create a list to contain the values to be sent back
		 * to Python.
		 */

		result = PyList_New( element_count );

		if ( result == NULL )
			return NULL;

		switch( field_type ) {
		case MX_CA_CHAR:
			char_array = (char *) value_ptr;
			break;
		case MX_CA_SHORT:
		case MX_CA_ENUM:
			short_array = (short *) value_ptr;
			break;
		case MX_CA_LONG:
			long_array = (long *) value_ptr;
			break;
		case MX_CA_FLOAT:
			float_array = (float *) value_ptr;
			break;
		case MX_CA_DOUBLE:
			double_array = (double *) value_ptr;
			break;
		}

		for ( i = 0; i < element_count; i++ ) {
			switch( field_type ) {
			case MX_CA_CHAR:
				item = Py_BuildValue( "b", char_array[i] );
				break;
			case MX_CA_SHORT:
			case MX_CA_ENUM:
				item = Py_BuildValue( "h", short_array[i] );
				break;
			case MX_CA_LONG:
				item = Py_BuildValue( "l", long_array[i] );
				break;
			case MX_CA_FLOAT:
				item = Py_BuildValue( "f", float_array[i] );
				break;
			case MX_CA_DOUBLE:
				item = Py_BuildValue( "d", double_array[i] );
				break;
			}

			python_status = PyList_SetItem( result, i, item );

			if ( python_status != 0 )
				return NULL;
		}
	}

	return result;
}

/*---------------------------------------------------------------------------*/

static long
mpca_py_string_as_long( PyObject *string_object )
{
	static const char fname[] = "mpca_py_string_as_long()";

	char *string_ptr;
	int num_items;
	long long_value;

	if ( string_object == (PyObject *) NULL ) {
		(void) mx_error( MXE_NULL_ARGUMENT, fname,
			"The Python object passed was NULL." );
		return (-1);
	}

	string_ptr = PyString_AsString( string_object );

	if ( string_ptr == NULL ) {
		(void) mx_error( MXE_TYPE_MISMATCH, fname,
			"Python object %p is not a string object.",
			string_object );
		return (-1);
	}

	num_items = sscanf( string_ptr, "%ld", &long_value );

	if ( num_items != 1 ) {
		(void) mx_error( MXE_TYPE_MISMATCH, fname,
		"The string '%s' from Python string object %p "
		"does not contain a long integer.",
			string_ptr, string_object );
		return (-1);
	}

	return long_value;
}

static double
mpca_py_string_as_double( PyObject *string_object )
{
	static const char fname[] = "mpca_py_string_as_double()";

	char *string_ptr;
	int num_items;
	double double_value;

	if ( string_object == (PyObject *) NULL ) {
		(void) mx_error( MXE_NULL_ARGUMENT, fname,
			"The Python object passed was NULL." );
		return (-1);
	}

	string_ptr = PyString_AsString( string_object );

	if ( string_ptr == NULL ) {
		(void) mx_error( MXE_TYPE_MISMATCH, fname,
			"Python object %p is not a string object.",
			string_object );
		return (-1);
	}

	num_items = sscanf( string_ptr, "%lg", &double_value );

	if ( num_items != 1 ) {
		(void) mx_error( MXE_TYPE_MISMATCH, fname,
		"The string '%s' from Python string object %p "
		"does not contain a floating point number.",
			string_ptr, string_object );
		return (-1);
	}

	return double_value;
}

/*---------------------------------------------------------------------------*/

MX_EXPORT PyObject *
MpCaPrivate_pv_caput_cmd( PyObject *self, PyObject *args )
{
	static const char fname[] = "MpCaPrivate_pv_caput_cmd()";

	PyObject *args_object, *sequence, *item;
	MX_EPICS_PV *pv;
	void *void_pv, *value_ptr;
	long pv_handle, field_type, wait;
	unsigned long i, length, element_count;
	unsigned long element_size, pv_max_bytes;
	double timeout;
	int parse_status;
	long long_value;
	double double_value;
	char *python_string_ptr;
	char *string_ptr = NULL;
	char *char_array = NULL;
	short *short_array = NULL;
	long *long_array = NULL;
	float *float_array = NULL;
	double *double_array = NULL;
	mx_bool_type decref_sequence;
	mx_status_type mx_status;

	MPCA_SHOW_THREAD_ID(fname);

#if 0
	mx_breakpoint();
#endif

	parse_status = PyArg_ParseTuple( args, "ldlO",
					&pv_handle, &timeout,
					&wait, &args_object );

	if ( parse_status == 0 )
		return NULL;

	/* Get the MX_EPICS_PV structure from the handle. */

	mx_status = mx_get_pointer_from_handle( &void_pv,
					MpCaPrivate_global_pv_handle_table,
					pv_handle );

	if ( mx_status.code != MXE_SUCCESS )
		return MpError_SetFromMxStatus( NULL, mx_status );

	pv = void_pv;

	/* Get the PV field type and length. */

	field_type = mx_epics_pv_get_field_type( pv );

	if ( field_type < 0 ) {
		mx_status = mx_error( MXE_NETWORK_IO_ERROR, fname,
		"Unable to get the native field type for EPICS PV '%s'.",
			pv->pvname );

		return MpError_SetFromMxStatus( NULL, mx_status );
	}

	element_size = mx_epics_get_element_size( field_type );

	element_count = mx_epics_pv_get_element_count( pv );

	if ( element_count == 0 ) {
		mx_status = mx_error( MXE_NETWORK_IO_ERROR, fname,
		"Unable to get the element count for EPICS PV '%s'.",
			pv->pvname );

		return MpError_SetFromMxStatus( NULL, mx_status );
	}

	pv_max_bytes = element_count * element_size;

	value_ptr = pv->application_ptr;

	switch( field_type ) {
	case MX_CA_STRING:
		string_ptr = (char *) value_ptr;
		break;
	case MX_CA_CHAR:
		char_array = (char *) value_ptr;
		break;
	case MX_CA_SHORT:
	case MX_CA_ENUM:
		short_array = (short *) value_ptr;
		break;
	case MX_CA_LONG:
		long_array = (long *) value_ptr;
		break;
	case MX_CA_FLOAT:
		float_array = (float *) value_ptr;
		break;
	case MX_CA_DOUBLE:
		double_array = (double *) value_ptr;
		break;
	}

	if ( PyList_Check( args_object ) ) {
		decref_sequence = FALSE;
		sequence = args_object;
	} else
	if ( PyTuple_Check( args_object ) ) {
		decref_sequence = FALSE;
		sequence = args_object;
	} else {
		decref_sequence = TRUE;

		/* If 'args_object' is not already a list or a tuple,
		 * then put it in a one-element tuple.
		 */

		sequence = PyTuple_New(1);

		if ( sequence == NULL )
			return NULL;

		PyTuple_SetItem( sequence, 0, args_object );
	}

	/* Walk through the sequence, converting Python objects to value
	 * that will be sent to EPICS.
	 */

	length = PySequence_Size( sequence );

	if ( length > element_count ) {
		length = element_count;
	}

	for ( i = 0; i < length; i++ ) {
		item = PySequence_GetItem( sequence, i );

		if ( item == NULL ) {
			if ( decref_sequence ) {
				Py_DECREF( sequence );
			}
			return NULL;
		}

		if ( PyString_Check( item ) ) {
			/* The item is a string. */

			switch( field_type ) {
			case MX_CA_STRING:
				python_string_ptr = PyString_AsString( item );

				strlcpy( string_ptr, python_string_ptr,
							pv_max_bytes );
				break;
			case MX_CA_CHAR:
				char_array[i] = (char)
					mpca_py_string_as_long( item );
				break;
			case MX_CA_SHORT:
			case MX_CA_ENUM:
				short_array[i] = (short)
					mpca_py_string_as_long( item );
				break;
			case MX_CA_LONG:
				long_array[i] = mpca_py_string_as_long( item );
				break;
			case MX_CA_FLOAT:
				float_array[i] = (float)
					mpca_py_string_as_double( item );
				break;
			case MX_CA_DOUBLE:
				double_array[i] = 
					mpca_py_string_as_double( item );
				break;
			}
		} else
		if ( PyInt_Check( item ) || PyLong_Check( item ) ) {
			long_value = PyInt_AsLong( item );

			switch( field_type ) {
			case MX_CA_STRING:
				snprintf( string_ptr, pv_max_bytes,
					"%ld", long_value );
				break;
			case MX_CA_CHAR:
				char_array[i] = (char) long_value;
				break;
			case MX_CA_SHORT:
			case MX_CA_ENUM:
				short_array[i] = (short) long_value;
				break;
			case MX_CA_LONG:
				long_array[i] = long_value;
				break;
			case MX_CA_FLOAT:
				float_array[i] = (float) long_value;
				break;
			case MX_CA_DOUBLE:
				double_array[i] = (double) long_value;
				break;
			}
		} else
		if ( PyFloat_Check( item ) ) {
			double_value = PyFloat_AsDouble( item );

			switch( field_type ) {
			case MX_CA_STRING:
				snprintf( string_ptr, pv_max_bytes,
					"%g", double_value );
				break;
			case MX_CA_CHAR:
				char_array[i] = (char) double_value;
				break;
			case MX_CA_SHORT:
			case MX_CA_ENUM:
				short_array[i] = (short) double_value;
				break;
			case MX_CA_LONG:
				long_array[i] = (long) double_value;
				break;
			case MX_CA_FLOAT:
				float_array[i] = (float) double_value;
				break;
			case MX_CA_DOUBLE:
				double_array[i] = double_value;
				break;
			}
		} else {
			PyObject *repr_object;

			repr_object = PyObject_Repr( item );

			mx_status = mx_error( MXE_ILLEGAL_ARGUMENT, fname,
			"Item %lu ('%s') passed is not a Python string, "
			"int, long, or double object.",
				i, PyString_AsString( repr_object ) );

			Py_DECREF( repr_object );

			if ( decref_sequence ) {
				Py_DECREF( sequence );
			}

			return MpError_SetFromMxStatus( NULL, mx_status );
		}
	}

	if ( decref_sequence ) {
		Py_DECREF( sequence );
	}

	if ( wait ) {
		mx_status = mx_caput( pv, field_type, length, value_ptr );
	} else
	if ( timeout > 0 ) {
		mx_status = mx_caput_with_timeout( pv, field_type,
						length, value_ptr, timeout );
	} else {
		mx_status = mx_caput_nowait( pv, field_type, length, value_ptr);
	}

	if ( mx_status.code != MXE_SUCCESS )
		return MpError_SetFromMxStatus( NULL, mx_status );

	Py_INCREF(Py_None);

	return Py_None;
}

/*---------------------------------------------------------------------------*/

#define MPCA_PRIVATE_CALLBACK_CLEANUP_AND_RELEASE_LOCK \
	do { 							\
		PyErr_WriteUnraisable( callback_object );	\
		PyErr_Clear();					\
		PyGILState_Release( python_GIL_state );		\
	} while(0)

static mx_status_type
MpCaPrivate_callback_function( MX_EPICS_CALLBACK *mx_epics_callback,
					void *arguments )
{
	static const char fname[] = "MpCaPrivate_callback_function()";

	PyObject *callback_object;
	PyObject *callback_function_object;
	PyObject *callback_arguments_object;
	PyObject *result;
	PyObject *arglist;
	PyGILState_STATE python_GIL_state;

	MX_EPICS_PV *pv;
	long field_type;
	unsigned long pv_element_size, pv_num_elements, pv_max_bytes;
	unsigned long server_element_size, num_server_elements;
	unsigned long server_bytes_sent, bytes_to_copy;
	const void *src_ptr;
	void *dest_ptr;

	MPCA_SHOW_THREAD_ID(fname);

#if MPCA_DEBUG
	MX_DEBUG(-2,("%s invoked for callback = %p, object = %p",
		fname, mx_epics_callback, callback_object ));
#endif

	pv = mx_epics_callback->pv;

	field_type = mx_epics_pv_get_field_type( pv );

	if ( field_type != mx_epics_callback->epics_type ) {
		return mx_error( MXE_UNSUPPORTED, fname,
		"The field type %ld received for EPICS PV '%s' does not "
		"match the local field type %ld.",
			mx_epics_callback->epics_type,
			pv->pvname,
			field_type );
	}

	/*--- Figure out the maximum number of bytes for this PV. ---*/

	pv_element_size = mx_epics_get_element_size( field_type );

	pv_num_elements = mx_epics_pv_get_element_count( pv );

	pv_max_bytes = pv_num_elements * pv_element_size;

	/*--- Figure out how many bytes were sent by the EPICS server. ---*/

	server_element_size = 
		mx_epics_get_element_size( mx_epics_callback->epics_type );

	num_server_elements = mx_epics_callback->epics_count;

	server_bytes_sent = num_server_elements * server_element_size;

	if ( server_bytes_sent <= pv_max_bytes ) {
		bytes_to_copy = server_bytes_sent;
	} else {
		bytes_to_copy = pv_max_bytes;

		mx_warning( "The %ld byte value received for PV '%s' "
			"will be truncated to %ld bytes.",
			server_bytes_sent, pv->pvname, bytes_to_copy );
	}

	/*--- Copy the received data to the PV. ---*/

	src_ptr = mx_epics_callback->value_ptr;

	dest_ptr = pv->application_ptr;

	if ( dest_ptr == src_ptr ) {
		mx_warning( "Destination pointer %p for EPICS PV '%s' "
		"is the same as the source pointer, so no copy "
		"will be performed.", dest_ptr, pv->pvname );
	} else {
		memcpy( dest_ptr, src_ptr, bytes_to_copy );
	}

	/* Acquire the Python Global Interpreter Lock.
	 * 
	 * Note: PyGILState_Ensure() and PyGILState_Release()
	 *       only exist for Python 2.3 and above.
	 */

	python_GIL_state = PyGILState_Ensure();

	/* Find the Python callback function. */

	callback_object = arguments;

	callback_function_object = PyObject_GetAttrString( callback_object,
							"callback_function" );

	if ( callback_function_object == NULL ) {
		MPCA_PRIVATE_CALLBACK_CLEANUP_AND_RELEASE_LOCK;

		return mx_error( MXE_CORRUPT_DATA_STRUCTURE | MXE_QUIET, fname,
		"The callback object passed did not have 'callback_function' "
		"as an attribute." );
	}

	callback_arguments_object = PyObject_GetAttrString( callback_object,
							"callback_arguments" );

	if ( callback_arguments_object == NULL ) {
		MPCA_PRIVATE_CALLBACK_CLEANUP_AND_RELEASE_LOCK;

		return mx_error( MXE_CORRUPT_DATA_STRUCTURE | MXE_QUIET, fname,
		"The callback object passed did not have 'callback_arguments' "
		"as an attribute." );
	}

	arglist = Py_BuildValue("(OO)",
			callback_object, callback_arguments_object );

	if ( arglist == NULL ) {
		MPCA_PRIVATE_CALLBACK_CLEANUP_AND_RELEASE_LOCK;

		return mx_error( MXE_FUNCTION_FAILED | MXE_QUIET, fname,
		"Wrapping the callback arguments in a tuple failed." );
	}

	/* Call the Python callback function. */

	result = PyEval_CallObject( callback_function_object, arglist );

	Py_DECREF( arglist );

	if ( result == NULL ) {
		MPCA_PRIVATE_CALLBACK_CLEANUP_AND_RELEASE_LOCK;

		return mx_error( MXE_FUNCTION_FAILED | MXE_QUIET, fname,
		"Calling the Python callback failed." );
	}

	Py_DECREF( result );

	PyGILState_Release( python_GIL_state );

	return MX_SUCCESSFUL_RESULT;
}

/*---------------------------------------------------------------------------*/

MX_EXPORT PyObject *
MpCaPrivate_pv_add_callback_cmd( PyObject *self, PyObject *args )
{
#if MPCA_DEBUG
	static const char fname[] = "MpCaPrivate_pv_add_callback_cmd()";
#endif

	PyObject *py_callback_object;
	PyObject *callback_mask_attribute, *callback_address_attribute;
	MX_EPICS_PV *pv;
	void *void_pv;
	MX_EPICS_CALLBACK *mx_callback_object;
	long pv_handle;
	unsigned long mx_callback_mask;
	int python_status;
	mx_status_type mx_status;

	MPCA_SHOW_THREAD_ID(fname);

#if 0
	mx_breakpoint();
#endif

	python_status = PyArg_ParseTuple( args, "lO", &pv_handle,
					&py_callback_object );

	if ( python_status == 0 )
		return NULL;

	/* Get the MX_EPICS_PV structure from the handle. */

	mx_status = mx_get_pointer_from_handle( &void_pv,
					MpCaPrivate_global_pv_handle_table,
					pv_handle );

	if ( mx_status.code != MXE_SUCCESS )
		return MpError_SetFromMxStatus( NULL, mx_status );

	pv = void_pv;

#if MPCA_DEBUG
	MX_DEBUG(-2,("%s invoked for PV '%s'", fname, pv->pvname ));
#endif

	/* Adding a callback will result in the Python callback function
	 * being invoked by EPICS from a thread that is not the main thread.
	 * Before we do this, we must check to see whether or not the Python
	 * Global Interpreter Lock has been initialized.  If it has not been
	 * initialized, we call PyEval_InitThreads() in the main thread to
	 * initialize it.
	 */

	if ( PyEval_ThreadsInitialized() == 0 ) {
		PyEval_InitThreads();
	}

	/* Get the callback mask from the callback object. */

	callback_mask_attribute = PyObject_GetAttrString( py_callback_object,
							"callback_mask" );

	if ( callback_mask_attribute == NULL )
		return NULL;

	mx_callback_mask = PyLong_AsLong( callback_mask_attribute );

	Py_DECREF( callback_mask_attribute );

	/* Setup the callback in the C library. */

	mx_status = mx_epics_add_callback( pv, mx_callback_mask,
					MpCaPrivate_callback_function,
					py_callback_object,
					&mx_callback_object );

	if ( mx_status.code != MXE_SUCCESS )
		return MpError_SetFromMxStatus( NULL, mx_status );

	/* Add the address of the C MX_CALLBACK structure as
	 * a private attribute.
	 */

	/* FIXME: Casting a pointer to an unsigned long is BAD! */

#if MPCA_DEBUG
	MX_DEBUG(-2,("%s: MX callback object = %p", fname, mx_callback_object));
#endif

	callback_address_attribute = 
		PyLong_FromLong((unsigned long) mx_callback_object );

	if ( callback_address_attribute == NULL )
		return NULL;

	python_status = PyObject_SetAttrString( py_callback_object,
						"__callback_address",
						callback_address_attribute );

	Py_DECREF( callback_address_attribute );

	if ( python_status == (-1) )
		return NULL;

	/* We are done, so return now. */

	Py_INCREF(Py_None);

	return Py_None;
}

/*---------------------------------------------------------------------------*/

MX_EXPORT PyObject *
MpCaPrivate_cb_delete_callback_cmd( PyObject *self, PyObject *args )
{
#if MPCA_DEBUG
	static const char fname[] = "MpCaPrivate_cb_delete_callback_cmd()";
#endif


	PyObject *py_callback_object, *callback_address_attribute;
	MX_EPICS_CALLBACK *mx_callback_object;
	int python_status;
	mx_status_type mx_status;

	MPCA_SHOW_THREAD_ID(fname);

#if 0
	mx_breakpoint();
#endif

	python_status = PyArg_ParseTuple( args, "O", &py_callback_object );

	if ( python_status == 0 )
		return NULL;

	callback_address_attribute = PyObject_GetAttrString( py_callback_object,
							"__callback_address" );

	if ( callback_address_attribute == NULL )
		return NULL;

	/* FIXME: Casting a long to a pointer is BAD! */

	mx_callback_object = (void *) 
			PyLong_AsLong( callback_address_attribute );

#if MPCA_DEBUG
	MX_DEBUG(-2,("%s: mx_callback_object = %p", fname, mx_callback_object));
#endif

	mx_status = mx_epics_delete_callback( mx_callback_object );

	if ( mx_status.code != MXE_SUCCESS )
		return MpError_SetFromMxStatus( NULL, mx_status );

	Py_INCREF( Py_None );

	return Py_None;
}

